@@ -1,6 +1,6 @@
 # MIT License
 #
-# Copyright (C) IBM Corporation 2019
+# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2019
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 # documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
@@ -24,12 +24,15 @@ predictions.
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 import logging
+from typing import Optional
 
 import numpy as np
+from tqdm import tqdm
 
-from art.config import ART_NUMPY_DTYPE
-from art.classifiers.classifier import Classifier
 from art.attacks.attack import EvasionAttack
+from art.config import ART_NUMPY_DTYPE
+from art.estimators.estimator import BaseEstimator
+from art.estimators.classification.classifier import Classifier, ClassifierMixin
 from art.utils import compute_success, to_categorical, check_and_transform_label_format
 
 logger = logging.getLogger(__name__)
@@ -55,94 +58,73 @@ class BoundaryAttack(EvasionAttack):
         "batch_size",
     ]
 
+    _estimator_requirements = (BaseEstimator, ClassifierMixin)
+
     def __init__(
         self,
-        classifier,
-        targeted=True,
-        delta=0.01,
-        epsilon=0.01,
-        step_adapt=0.667,
-        max_iter=5000,
-        num_trial=25,
-        sample_size=20,
-        init_size=100,
-    ):
+        estimator: Classifier,
+        targeted: bool = True,
+        delta: float = 0.01,
+        epsilon: float = 0.01,
+        step_adapt: float = 0.667,
+        max_iter: int = 5000,
+        num_trial: int = 25,
+        sample_size: int = 20,
+        init_size: int = 100,
+    ) -> None:
         """
         Create a boundary attack instance.
 
-        :param classifier: A trained classifier.
-        :type classifier: :class:`.Classifier`
+        :param estimator: A trained classifier.
         :param targeted: Should the attack target one specific class.
-        :type targeted: `bool`
         :param delta: Initial step size for the orthogonal step.
-        :type delta: `float`
         :param epsilon: Initial step size for the step towards the target.
-        :type epsilon: `float`
         :param step_adapt: Factor by which the step sizes are multiplied or divided, must be in the range (0, 1).
-        :type step_adapt: `float`
         :param max_iter: Maximum number of iterations.
-        :type max_iter: `int`
         :param num_trial: Maximum number of trials per iteration.
-        :type num_trial: `int`
         :param sample_size: Number of samples per trial.
-        :type sample_size: `int`
         :param init_size: Maximum number of trials for initial generation of adversarial examples.
-        :type init_size: `int`
         """
-        super(BoundaryAttack, self).__init__(classifier=classifier)
+        super(BoundaryAttack, self).__init__(estimator=estimator)
 
-        params = {
-            "targeted": targeted,
-            "delta": delta,
-            "epsilon": epsilon,
-            "step_adapt": step_adapt,
-            "max_iter": max_iter,
-            "num_trial": num_trial,
-            "sample_size": sample_size,
-            "init_size": init_size,
-            "batch_size": 1,
-        }
-        self.set_params(**params)
+        self.targeted = targeted
+        self.delta = delta
+        self.epsilon = epsilon
+        self.step_adapt = step_adapt
+        self.max_iter = max_iter
+        self.num_trial = num_trial
+        self.sample_size = sample_size
+        self.init_size = init_size
+        self.batch_size = 1
+        self._check_params()
 
-    @classmethod
-    def is_valid_classifier_type(cls, classifier):
-        """
-        Checks whether the classifier provided is a classifer which this class can perform an attack on
-        :param classifier:
-        :return:
-        """
-        return True if isinstance(classifier, Classifier) else False
-
-    def generate(self, x, y=None, **kwargs):
+    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
         """
         Generate adversarial samples and return them in an array.
 
         :param x: An array with the original inputs to be attacked.
-        :type x: `np.ndarray`
         :param y: Target values (class labels) one-hot-encoded of shape (nb_samples, nb_classes) or indices of shape
                   (nb_samples,). If `self.targeted` is true, then `y` represents the target labels.
-        :type y: `np.ndarray` or `None`
         :param x_adv_init: Initial array to act as initial adversarial examples. Same shape as `x`.
         :type x_adv_init: `np.ndarray`
         :return: An array holding the adversarial examples.
-        :rtype: `np.ndarray`
         """
-        y = check_and_transform_label_format(y, self.classifier.nb_classes(), return_one_hot=False)
+        y = check_and_transform_label_format(y, self.estimator.nb_classes, return_one_hot=False)
 
         # Get clip_min and clip_max from the classifier or infer them from data
-        if hasattr(self.classifier, "clip_values") and self.classifier.clip_values is not None:
-            clip_min, clip_max = self.classifier.clip_values
+        if self.estimator.clip_values is not None:
+            clip_min, clip_max = self.estimator.clip_values
         else:
             clip_min, clip_max = np.min(x), np.max(x)
 
         # Prediction from the original images
-        preds = np.argmax(self.classifier.predict(x, batch_size=self.batch_size), axis=1)
+        preds = np.argmax(self.estimator.predict(x, batch_size=self.batch_size), axis=1)
 
         # Prediction from the initial adversarial examples if not None
         x_adv_init = kwargs.get("x_adv_init")
 
         if x_adv_init is not None:
-            init_preds = np.argmax(self.classifier.predict(x_adv_init, batch_size=self.batch_size), axis=1)
+            init_preds = np.argmax(self.estimator.predict(x_adv_init, batch_size=self.batch_size), axis=1)
         else:
             init_preds = [None] * len(x)
             x_adv_init = [None] * len(x)
@@ -155,7 +137,7 @@ class BoundaryAttack(EvasionAttack):
         x_adv = x.astype(ART_NUMPY_DTYPE)
 
         # Generate the adversarial samples
-        for ind, val in enumerate(x_adv):
+        for ind, val in enumerate(tqdm(x_adv, desc="Boundary attack")):
             if self.targeted:
                 x_adv[ind] = self._perturb(
                     x=val,
@@ -178,35 +160,29 @@ class BoundaryAttack(EvasionAttack):
                 )
 
         if y is not None:
-            y = to_categorical(y, self.classifier.nb_classes())
+            y = to_categorical(y, self.estimator.nb_classes)
 
         logger.info(
             "Success rate of Boundary attack: %.2f%%",
-            100 * compute_success(self.classifier, x, y, x_adv, self.targeted, batch_size=self.batch_size),
+            100 * compute_success(self.estimator, x, y, x_adv, self.targeted, batch_size=self.batch_size),
         )
 
         return x_adv
 
-    def _perturb(self, x, y, y_p, init_pred, adv_init, clip_min, clip_max):
+    def _perturb(
+        self, x: np.ndarray, y: int, y_p: int, init_pred: int, adv_init: np.ndarray, clip_min: float, clip_max: float,
+    ) -> np.ndarray:
         """
         Internal attack function for one example.
 
         :param x: An array with one original input to be attacked.
-        :type x: `np.ndarray`
         :param y: If `self.targeted` is true, then `y` represents the target label.
-        :type y: `int`
         :param y_p: The predicted label of x.
-        :type y_p: `int`
         :param init_pred: The predicted label of the initial image.
-        :type init_pred: `int`
         :param adv_init: Initial array to act as an initial adversarial example.
-        :type adv_init: `np.ndarray`
         :param clip_min: Minimum value of an example.
-        :type clip_min: `float`
         :param clip_max: Maximum value of an example.
-        :type clip_max: `float`
-        :return: an adversarial example.
-        :rtype: `np.ndarray`
+        :return: An adversarial example.
         """
         # First, create an initial adversarial sample
         initial_sample = self._init_sample(x, y, y_p, init_pred, adv_init, clip_min, clip_max)
@@ -216,30 +192,31 @@ class BoundaryAttack(EvasionAttack):
             return x
 
         # If an initial adversarial example found, then go with boundary attack
-        x_adv = self._attack(initial_sample[0], x, initial_sample[1], self.delta, self.epsilon, clip_min, clip_max)
+        x_adv = self._attack(initial_sample[0], x, initial_sample[1], self.delta, self.epsilon, clip_min, clip_max,)
 
         return x_adv
 
-    def _attack(self, initial_sample, original_sample, target, initial_delta, initial_epsilon, clip_min, clip_max):
+    def _attack(
+        self,
+        initial_sample: np.ndarray,
+        original_sample: np.ndarray,
+        target: int,
+        initial_delta: float,
+        initial_epsilon: float,
+        clip_min: float,
+        clip_max: float,
+    ) -> np.ndarray:
         """
         Main function for the boundary attack.
 
         :param initial_sample: An initial adversarial example.
-        :type initial_sample: `np.ndarray`
         :param original_sample: The original input.
-        :type original_sample: `np.ndarray`
         :param target: The target label.
-        :type target: `int`
         :param initial_delta: Initial step size for the orthogonal step.
-        :type initial_delta: `float`
         :param initial_epsilon: Initial step size for the step towards the target.
-        :type initial_epsilon: `float`
         :param clip_min: Minimum value of an example.
-        :type clip_min: `float`
         :param clip_max: Maximum value of an example.
-        :type clip_max: `float`
         :return: an adversarial example.
-        :rtype: `np.ndarray`
         """
         # Get initialization for some variables
         x_adv = initial_sample
@@ -256,7 +233,7 @@ class BoundaryAttack(EvasionAttack):
                     potential_adv = np.clip(potential_adv, clip_min, clip_max)
                     potential_advs.append(potential_adv)
 
-                preds = np.argmax(self.classifier.predict(np.array(potential_advs), batch_size=self.batch_size), axis=1)
+                preds = np.argmax(self.estimator.predict(np.array(potential_advs), batch_size=self.batch_size), axis=1,)
                 satisfied = preds == target
                 delta_ratio = np.mean(satisfied)
 
@@ -278,7 +255,7 @@ class BoundaryAttack(EvasionAttack):
                 perturb *= self.curr_epsilon
                 potential_advs = x_advs + perturb
                 potential_advs = np.clip(potential_advs, clip_min, clip_max)
-                preds = np.argmax(self.classifier.predict(potential_advs, batch_size=self.batch_size), axis=1)
+                preds = np.argmax(self.estimator.predict(potential_advs, batch_size=self.batch_size), axis=1,)
                 satisfied = preds == target
                 epsilon_ratio = np.mean(satisfied)
 
@@ -296,21 +273,17 @@ class BoundaryAttack(EvasionAttack):
 
         return x_adv
 
-    def _orthogonal_perturb(self, delta, current_sample, original_sample):
+    def _orthogonal_perturb(self, delta: float, current_sample: np.ndarray, original_sample: np.ndarray) -> np.ndarray:
         """
         Create an orthogonal perturbation.
 
         :param delta: Initial step size for the orthogonal step.
-        :type delta: `float`
         :param current_sample: Current adversarial example.
-        :type current_sample: `np.ndarray`
         :param original_sample: The original input.
-        :type original_sample: `np.ndarray`
         :return: a possible perturbation.
-        :rtype: `np.ndarray`
         """
         # Generate perturbation randomly
-        perturb = np.random.randn(*self.classifier.input_shape).astype(ART_NUMPY_DTYPE)
+        perturb = np.random.randn(*self.estimator.input_shape).astype(ART_NUMPY_DTYPE)
 
         # Rescale the perturbation
         perturb /= np.linalg.norm(perturb)
@@ -319,42 +292,38 @@ class BoundaryAttack(EvasionAttack):
         # Project the perturbation onto sphere
         direction = original_sample - current_sample
 
-        if len(self.classifier.input_shape) == 3:
-            perturb = np.swapaxes(perturb, 0, self.classifier.channel_index - 1)
-            direction = np.swapaxes(direction, 0, self.classifier.channel_index - 1)
+        if len(self.estimator.input_shape) == 3:
+            channel_index = 1 if self.estimator.channels_first else 3
+            perturb = np.swapaxes(perturb, 0, channel_index - 1)
+            direction = np.swapaxes(direction, 0, channel_index - 1)
             for i in range(direction.shape[0]):
                 direction[i] /= np.linalg.norm(direction[i])
                 perturb[i] -= np.dot(np.dot(perturb[i], direction[i].T), direction[i])
 
-            perturb = np.swapaxes(perturb, 0, self.classifier.channel_index - 1)
-        elif len(self.classifier.input_shape) == 1:
+            perturb = np.swapaxes(perturb, 0, channel_index - 1)
+        elif len(self.estimator.input_shape) == 1:
             direction /= np.linalg.norm(direction)
             perturb -= np.dot(perturb, direction.T) * direction
         else:
             raise ValueError("Input shape not recognised.")
-
+        hypotenuse = np.sqrt(1 + delta ** 2)
+        perturb = ((1 - hypotenuse) * (current_sample - original_sample) + perturb) / hypotenuse
         return perturb
 
-    def _init_sample(self, x, y, y_p, init_pred, adv_init, clip_min, clip_max):
+    def _init_sample(
+        self, x: np.ndarray, y: int, y_p: int, init_pred: int, adv_init: np.ndarray, clip_min: float, clip_max: float,
+    ) -> Optional[np.ndarray]:
         """
         Find initial adversarial example for the attack.
 
-        :param x: An array with 1 original input to be attacked.
-        :type x: `np.ndarray`
+        :param x: An array with one original input to be attacked.
         :param y: If `self.targeted` is true, then `y` represents the target label.
-        :type y: `int`
         :param y_p: The predicted label of x.
-        :type y_p: `int`
         :param init_pred: The predicted label of the initial image.
-        :type init_pred: `int`
         :param adv_init: Initial array to act as an initial adversarial example.
-        :type adv_init: `np.ndarray`
         :param clip_min: Minimum value of an example.
-        :type clip_min: `float`
         :param clip_max: Maximum value of an example.
-        :type clip_max: `float`
         :return: an adversarial example.
-        :rtype: `np.ndarray`
         """
         nprd = np.random.RandomState()
         initial_sample = None
@@ -372,7 +341,7 @@ class BoundaryAttack(EvasionAttack):
             for _ in range(self.init_size):
                 random_img = nprd.uniform(clip_min, clip_max, size=x.shape).astype(x.dtype)
                 random_class = np.argmax(
-                    self.classifier.predict(np.array([random_img]), batch_size=self.batch_size), axis=1
+                    self.estimator.predict(np.array([random_img]), batch_size=self.batch_size), axis=1,
                 )[0]
 
                 if random_class == y:
@@ -392,7 +361,7 @@ class BoundaryAttack(EvasionAttack):
             for _ in range(self.init_size):
                 random_img = nprd.uniform(clip_min, clip_max, size=x.shape).astype(x.dtype)
                 random_class = np.argmax(
-                    self.classifier.predict(np.array([random_img]), batch_size=self.batch_size), axis=1
+                    self.estimator.predict(np.array([random_img]), batch_size=self.batch_size), axis=1,
                 )[0]
 
                 if random_class != y_p:
@@ -405,30 +374,7 @@ class BoundaryAttack(EvasionAttack):
 
         return initial_sample
 
-    def set_params(self, **kwargs):
-        """
-        Take in a dictionary of parameters and applies attack-specific checks before saving them as attributes.
-
-        :param targeted: Should the attack target one specific class.
-        :type targeted: `bool`
-        :param delta: Initial step size for the orthogonal step.
-        :type delta: `float`
-        :param epsilon: Initial step size for the step towards the target.
-        :type epsilon: `float`
-        :param step_adapt: Factor by which the step sizes are multiplied or divided, must be in the range (0, 1).
-        :type step_adapt: `float`
-        :param max_iter: Maximum number of iterations.
-        :type max_iter: `int`
-        :param num_trial: Maximum number of trials per iteration.
-        :type num_trial: `int`
-        :param sample_size: Number of samples per trial.
-        :type sample_size: `int`
-        :param init_size: Maximum number of trials for initial generation of adversarial examples.
-        :type init_size: `int`
-        """
-        # Save attack-specific parameters
-        super(BoundaryAttack, self).set_params(**kwargs)
-
+    def _check_params(self) -> None:
         if not isinstance(self.max_iter, (int, np.int)) or self.max_iter < 0:
             raise ValueError("The number of iterations must be a non-negative integer.")
 
@@ -449,5 +395,3 @@ class BoundaryAttack(EvasionAttack):
 
         if self.step_adapt <= 0 or self.step_adapt >= 1:
             raise ValueError("The adaptation factor must be in the range (0, 1).")
-
-        return True
