@@ -1,6 +1,6 @@
 # MIT License
 #
-# Copyright (C) IBM Corporation 2018
+# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2018
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 # documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
@@ -21,11 +21,13 @@ This module implements attacks on Decision Trees.
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 import logging
+from typing import List, Optional, Union
 
 import numpy as np
+from tqdm import trange
 
 from art.attacks.attack import EvasionAttack
-from art.classifiers.scikitlearn import ScikitlearnDecisionTreeClassifier
+from art.estimators.classification.scikitlearn import ScikitlearnDecisionTreeClassifier
 from art.utils import check_and_transform_label_format, compute_success
 
 logger = logging.getLogger(__name__)
@@ -40,52 +42,46 @@ class DecisionTreeAttack(EvasionAttack):
     """
 
     attack_params = ["classifier", "offset"]
+    _estimator_requirements = (ScikitlearnDecisionTreeClassifier,)
 
-    def __init__(self, classifier, offset=0.001):
+    def __init__(self, classifier: ScikitlearnDecisionTreeClassifier, offset: float = 0.001) -> None:
         """
         :param classifier: A trained model of type scikit decision tree.
-        :type classifier: :class:`.Classifier.ScikitlearnDecisionTreeClassifier`
-        :param offset: How much the value is pushed away from tree's threshold. default 0.001
-        :type classifier: :float:
+        :param offset: How much the value is pushed away from tree's threshold.
         """
-        super(DecisionTreeAttack, self).__init__(classifier)
+        super(DecisionTreeAttack, self).__init__(estimator=classifier)
+        self.offset = offset
+        self._check_params()
 
-        if not isinstance(classifier, ScikitlearnDecisionTreeClassifier):
-            raise TypeError("Model must be a decision tree model!")
-        params = {"offset": offset}
-        self.set_params(**params)
-
-    def _df_subtree(self, position, original_class, target=None):
+    def _df_subtree(
+        self, position: int, original_class: Union[int, np.ndarray], target: Optional[int] = None,
+    ) -> List[int]:
         """
         Search a decision tree for a mis-classifying instance.
 
         :param position: An array with the original inputs to be attacked.
-        :type position: `int`
         :param original_class: original label for the instances we are searching mis-classification for.
-        :type original_class: `int` or `np.ndarray`
         :param target: If the provided, specifies which output the leaf has to have to be accepted.
-        :type target: `int`
         :return: An array specifying the path to the leaf where the classification is either != original class or
                  ==target class if provided.
-        :rtype: `list`
         """
         # base case, we're at a leaf
-        if self.classifier.get_left_child(position) == self.classifier.get_right_child(position):
+        if self.estimator.get_left_child(position) == self.estimator.get_right_child(position):
             if target is None:  # untargeted case
-                if self.classifier.get_classes_at_node(position) != original_class:
+                if self.estimator.get_classes_at_node(position) != original_class:
                     path = [position]
                 else:
                     path = [-1]
             else:  # targeted case
-                if self.classifier.get_classes_at_node(position) == target:
+                if self.estimator.get_classes_at_node(position) == target:
                     path = [position]
                 else:
                     path = [-1]
         else:  # go deeper, depths first
-            res = self._df_subtree(self.classifier.get_left_child(position), original_class, target)
+            res = self._df_subtree(self.estimator.get_left_child(position), original_class, target)
             if res[0] == -1:
                 # no result, try right subtree
-                res = self._df_subtree(self.classifier.get_right_child(position), original_class, target)
+                res = self._df_subtree(self.estimator.get_right_child(position), original_class, target)
                 if res[0] == -1:
                     # no desired result
                     path = [-1]
@@ -99,24 +95,21 @@ class DecisionTreeAttack(EvasionAttack):
 
         return path
 
-    def generate(self, x, y=None, **kwargs):
+    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
         """
         Generate adversarial examples and return them as an array.
 
         :param x: An array with the original inputs to be attacked.
-        :type x: `np.ndarray`
         :param y: Target values (class labels) one-hot-encoded of shape (nb_samples, nb_classes) or indices of shape
                   (nb_samples,).
-        :type y: `np.ndarray`
         :return: An array holding the adversarial examples.
-        :rtype: `np.ndarray`
         """
-        y = check_and_transform_label_format(y, self.classifier.nb_classes(), return_one_hot=False)
+        y = check_and_transform_label_format(y, self.estimator.nb_classes, return_one_hot=False)
         x_adv = x.copy()
 
-        for index in range(np.shape(x_adv)[0]):
-            path = self.classifier.get_decision_path(x_adv[index])
-            legitimate_class = np.argmax(self.classifier.predict(x_adv[index].reshape(1, -1)))
+        for index in trange(x_adv.shape[0], desc="Decision tree attack"):
+            path = self.estimator.get_decision_path(x_adv[index])
+            legitimate_class = np.argmax(self.estimator.predict(x_adv[index].reshape(1, -1)))
             position = -2
             adv_path = [-1]
             ancestor = path[position]
@@ -124,19 +117,19 @@ class DecisionTreeAttack(EvasionAttack):
                 ancestor = path[position]
                 current_child = path[position + 1]
                 # search in right subtree
-                if current_child == self.classifier.get_left_child(ancestor):
+                if current_child == self.estimator.get_left_child(ancestor):
                     if y is None:
-                        adv_path = self._df_subtree(self.classifier.get_right_child(ancestor), legitimate_class)
+                        adv_path = self._df_subtree(self.estimator.get_right_child(ancestor), legitimate_class)
                     else:
                         adv_path = self._df_subtree(
-                            self.classifier.get_right_child(ancestor), legitimate_class, y[index]
+                            self.estimator.get_right_child(ancestor), legitimate_class, y[index],
                         )
                 else:  # search in left subtree
                     if y is None:
-                        adv_path = self._df_subtree(self.classifier.get_left_child(ancestor), legitimate_class)
+                        adv_path = self._df_subtree(self.estimator.get_left_child(ancestor), legitimate_class)
                     else:
                         adv_path = self._df_subtree(
-                            self.classifier.get_left_child(ancestor), legitimate_class, y[index]
+                            self.estimator.get_left_child(ancestor), legitimate_class, y[index],
                         )
                 position = position - 1  # we are going the decision path upwards
             adv_path.append(ancestor)
@@ -144,26 +137,22 @@ class DecisionTreeAttack(EvasionAttack):
             # first one is leaf-> no threshold, cannot be perturbed
             for i in range(1, 1 + len(adv_path[1:])):
                 go_for = adv_path[i - 1]
-                threshold = self.classifier.get_threshold_at_node(adv_path[i])
-                feature = self.classifier.get_feature_at_node(adv_path[i])
+                threshold = self.estimator.get_threshold_at_node(adv_path[i])
+                feature = self.estimator.get_feature_at_node(adv_path[i])
                 # only perturb if the feature is actually wrong
-                if x_adv[index][feature] > threshold and go_for == self.classifier.get_left_child(adv_path[i]):
+                if x_adv[index][feature] > threshold and go_for == self.estimator.get_left_child(adv_path[i]):
                     x_adv[index][feature] = threshold - self.offset
-                elif x_adv[index][feature] <= threshold and go_for == self.classifier.get_right_child(adv_path[i]):
+                elif x_adv[index][feature] <= threshold and go_for == self.estimator.get_right_child(adv_path[i]):
                     x_adv[index][feature] = threshold + self.offset
 
-        logger.info("Success rate of decision tree attack: %.2f%%", 100 * compute_success(self.classifier, x, y, x_adv))
+        logger.info(
+            "Success rate of decision tree attack: %.2f%%", 100 * compute_success(self.estimator, x, y, x_adv),
+        )
         return x_adv
 
-    def set_params(self, **kwargs):
-        """
-        Take in a dictionary of parameters and apply attack-specific checks before saving them as attributes.
-
-        :param kwargs: a dictionary of attack-specific parameters
-        :type kwargs: `dict`
-        :return: `True` when parsing was successful
-        """
-        super(DecisionTreeAttack, self).set_params(**kwargs)
+    def _check_params(self) -> None:
+        if not isinstance(self.estimator, ScikitlearnDecisionTreeClassifier):
+            raise TypeError("Model must be a decision tree model.")
 
         if self.offset <= 0:
             raise ValueError("The offset parameter must be strictly positive.")
