@@ -1,6 +1,7 @@
 # MIT License
 #
-# Copyright (C) IBM Corporation 2019
+# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2019
+#
 # Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 # documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
 # rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
@@ -26,15 +27,21 @@ Extracted model - Test Fidelity: 0.9977
 
 | Paper link: https://arxiv.org/abs/1909.01838
 """
-
-import os
 import logging
+import os
+from typing import List, Optional, TYPE_CHECKING
 
 import numpy as np
 from scipy.optimize import least_squares
 
-from art.classifiers import KerasClassifier, BlackBoxClassifier
 from art.attacks.attack import ExtractionAttack
+from art.estimators.estimator import BaseEstimator, NeuralNetworkMixin
+from art.estimators.classification.classifier import ClassifierMixin
+from art.estimators.classification.keras import KerasClassifier
+from art.estimators.classification.blackbox import BlackBoxClassifier
+
+if TYPE_CHECKING:
+    from art.estimators.classification.classifier import Classifier
 
 NUMPY_DTYPE = np.float64
 
@@ -49,80 +56,66 @@ class FunctionallyEquivalentExtraction(ExtractionAttack):
     | Paper link: https://arxiv.org/abs/1909.01838
     """
 
-    def __init__(self, classifier, num_neurons):
+    _estimator_requirements = (BaseEstimator, NeuralNetworkMixin, ClassifierMixin)
+
+    def __init__(self, classifier: "Classifier", num_neurons: Optional[int] = None) -> None:
         """
         Create a `FunctionallyEquivalentExtraction` instance.
 
         :param classifier: A trained ART classifier.
-        :type classifier: :class:`.Classifier`
         :param num_neurons: The number of neurons in the first dense layer.
-        :type num_neurons: `int`
         """
-        super().__init__(classifier)
+        super().__init__(estimator=classifier)
         self.num_neurons = num_neurons
-        self.num_classes = classifier.nb_classes()
+        self.num_classes = classifier.nb_classes
         self.num_features = int(np.prod(classifier.input_shape))
 
         self.vector_u = np.random.normal(0, 1, (1, self.num_features)).astype(dtype=NUMPY_DTYPE)
         self.vector_v = np.random.normal(0, 1, (1, self.num_features)).astype(dtype=NUMPY_DTYPE)
 
-        self.critical_points = list()
+        self.critical_points: List[np.ndarray] = list()
 
-        self.w_0 = None  # Weight matrix of first dense layer
-        self.b_0 = None  # Bias vector of first dense layer
-        self.w_1 = None  # weight matrix of second dense layer
-        self.b_1 = None  # Bias vector of second dense layer
+        self.w_0: Optional[np.ndarray] = None  # Weight matrix of first dense layer
+        self.b_0: Optional[np.ndarray] = None  # Bias vector of first dense layer
+        self.w_1: Optional[np.ndarray] = None  # Weight matrix of second dense layer
+        self.b_1: Optional[np.ndarray] = None  # Bias vector of second dense layer
 
     def extract(
         self,
-        x,
-        y=None,
-        delta_0=0.05,
-        fraction_true=0.3,
-        rel_diff_slope=0.00001,
-        rel_diff_value=0.000001,
-        delta_init_value=0.1,
-        delta_value_max=50,
-        d2_min=0.0004,
-        d_step=0.01,
-        delta_sign=0.02,
-        unit_vector_scale=10000,
+        x: np.ndarray,
+        y: Optional[np.ndarray] = None,
+        delta_0: float = 0.05,
+        fraction_true: float = 0.3,
+        rel_diff_slope: float = 0.00001,
+        rel_diff_value: float = 0.000001,
+        delta_init_value: float = 0.1,
+        delta_value_max: int = 50,
+        d2_min: float = 0.0004,
+        d_step: float = 0.01,
+        delta_sign: float = 0.02,
+        unit_vector_scale: int = 10000,
         **kwargs
-    ):
+    ) -> BlackBoxClassifier:
         """
         Extract the targeted model.
 
         :param x: Samples of input data of shape (num_samples, num_features).
-        :type x: `np.ndarray`
         :param y: Correct labels or target labels for `x`, depending if the attack is targeted
                or not. This parameter is only used by some of the attacks.
-        :type y: `np.ndarray`
-        :param delta_0: Initial step size of binary search
-        :type delta_0: `float`
-        :param fraction_true: Fraction of output predictions that have to fulfill criteria for critical point
-        :type fraction_true: `float`
-        :param rel_diff_slope: Relative slope difference at critical points
-        :type rel_diff_slope: `float`
-        :param rel_diff_value: Relative value difference at critical points
-        :type rel_diff_value: `float`
-        :param delta_init_value: Initial delta of weight value search
-        :type delta_init_value: `float`
-        :param delta_value_max: Maximum delta  of weight value search
-        :type delta_value_max: `float`
-        :param d2_min: Minimum acceptable value of sum of absolute second derivatives
-        :type d2_min: `float`
-        :param d_step:  Step size of delta increase
-        :type d_step: `float`
-        :param delta_sign: Delta of weight sign search
-        :type delta_sign: `float`
-        :param unit_vector_scale: Multiplicative scale of the unit vector e_j.
-        :type unit_vector_scale: `int`
-
-        :return: ART BlackBoxClassifier of the extracted model.
-        :rtype: :class:`.BlackBoxClassifier`
+        :param delta_0: Initial step size of binary search.
+        :param fraction_true: Fraction of output predictions that have to fulfill criteria for critical point.
+        :param rel_diff_slope: Relative slope difference at critical points.
+        :param rel_diff_value: Relative value difference at critical points.
+        :param delta_init_value: Initial delta of weight value search.
+        :param delta_value_max: Maximum delta  of weight value search.
+        :param d2_min: Minimum acceptable value of sum of absolute second derivatives.
+        :param d_step:  Step size of delta increase.
+        :param delta_sign: Delta of weight sign search.
+        :param unit_vector_scale: Multiplicative scale of the unit vector `e_j`.
+        :return: ART :class:`.BlackBoxClassifier` of the extracted model.
         """
         self._critical_point_search(
-            delta_0=delta_0, fraction_true=fraction_true, rel_diff_slope=rel_diff_slope, rel_diff_value=rel_diff_value
+            delta_0=delta_0, fraction_true=fraction_true, rel_diff_slope=rel_diff_slope, rel_diff_value=rel_diff_value,
         )
         self._weight_recovery(
             delta_init_value=delta_init_value,
@@ -134,81 +127,72 @@ class FunctionallyEquivalentExtraction(ExtractionAttack):
         self._sign_recovery(unit_vector_scale=unit_vector_scale)
         self._last_layer_extraction(x)
 
-        def predict(x):
+        def predict(x: np.ndarray) -> np.ndarray:
             """
             Predict extracted model.
 
-            :param x: Samples of input data of shape (num_samples, num_features)
-            :type x: `np.ndarray`
-
-            :return: Predictions with the extracted model of shape (num_samples, num_classes)
-            :rtype: `np.ndarray`
+            :param x: Samples of input data of shape `(num_samples, num_features)`.
+            :return: Predictions with the extracted model of shape `(num_samples, num_classes)`.
             """
-            layer_0 = np.maximum(np.matmul(self.w_0.T, x.T) + self.b_0, 0.0)
-            layer_1 = np.matmul(self.w_1.T, layer_0) + self.b_1
+            layer_0 = np.maximum(np.matmul(self.w_0.T, x.T) + self.b_0, 0.0)  # type: ignore
+            layer_1 = np.matmul(self.w_1.T, layer_0) + self.b_1  # type: ignore
             return layer_1.T
 
         extracted_classifier = BlackBoxClassifier(
             predict,
-            input_shape=self.classifier.input_shape,
-            nb_classes=self.classifier.nb_classes(),
-            clip_values=self.classifier.clip_values,
-            preprocessing_defences=self.classifier.preprocessing_defences,
-            preprocessing=self.classifier.preprocessing,
+            input_shape=self.estimator.input_shape,
+            nb_classes=self.estimator.nb_classes,
+            clip_values=self.estimator.clip_values,
+            preprocessing_defences=self.estimator.preprocessing_defences,
+            preprocessing=self.estimator.preprocessing,
         )
 
         return extracted_classifier
 
-    def _o_l(self, x, e_j=None):
+    def _o_l(self, x: np.ndarray, e_j: Optional[np.ndarray] = None) -> np.ndarray:
         """
         Predict the target model.
 
-        :param x: Samples of input data of shape (num_samples, num_features)
-        :type x: `np.ndarray`
-        :param e_j: Additive delta vector of shape (1, num_features)
-        :type e_j: `np.ndarray`
-        :return: Prediction of the target model of shape (num_samples, num_classes)
-        :rtype: `np.ndarray`
+        :param x: Samples of input data of shape `(num_samples, num_features)`.
+        :param e_j: Additive delta vector of shape `(1, num_features)`.
+        :return: Prediction of the target model of shape `(num_samples, num_classes)`.
         """
         if e_j is not None:
             x = x + e_j
-        return self.classifier.predict(x).astype(NUMPY_DTYPE)
+        return self.estimator.predict(x).astype(NUMPY_DTYPE)
 
-    def _get_x(self, var_t):
+    def _get_x(self, var_t: float) -> np.ndarray:
         """
         Get input sample as function of multiplicative factor of random vector.
 
-        :param var_t: Multiplicative factor of second random vector for critical point search
-        :type var_t: `float`
-        :return: Input sample of shape (1, num_features)
-        :rtype: `np.ndarray`
+        :param var_t: Multiplicative factor of second random vector for critical point search.
+        :return: Input sample of shape `(1, num_features)`.
         """
         return self.vector_u + var_t * self.vector_v
 
-    def _critical_point_search(self, delta_0, fraction_true, rel_diff_slope, rel_diff_value):
+    def _critical_point_search(
+        self, delta_0: float, fraction_true: float, rel_diff_slope: float, rel_diff_value: float,
+    ) -> None:
         """
         Search for critical points.
 
-        :param delta_0: Initial step size of binary search
-        :type delta_0: `float`
-        :param fraction_true: Fraction of output predictions that have to fulfill criteria for critical point
-        :type fraction_true: `float`
-        :param rel_diff_slope: Relative slope difference at critical points
-        :type rel_diff_slope: `float`
-        :param rel_diff_value: Relative value difference at critical points
-        :type rel_diff_value: `float`
+        :param delta_0: Initial step size of binary search.
+        :param fraction_true: Fraction of output predictions that have to fulfill criteria for critical point.
+        :param rel_diff_slope: Relative slope difference at critical points.
+        :param rel_diff_value: Relative value difference at critical points.
         """
         logger.info("Searching for critical points.")
+
+        if self.num_neurons is None:
+            raise ValueError("The value of `num_neurons` is required for critical point search.")
         h_square = self.num_neurons * self.num_neurons
 
-        t_current = -h_square
+        t_current = float(-h_square)
         while t_current < h_square:
-
             delta = delta_0
             found_critical_point = False
 
             while not found_critical_point:
-
                 epsilon = delta / 10
 
                 t_1 = t_current
@@ -260,25 +244,24 @@ class FunctionallyEquivalentExtraction(ExtractionAttack):
                 "neurons in the first layer ({}).".format(len(self.critical_points), self.num_neurons)
             )
 
-    def _weight_recovery(self, delta_init_value, delta_value_max, d2_min, d_step, delta_sign):
+    def _weight_recovery(
+        self, delta_init_value: float, delta_value_max: float, d2_min: float, d_step: float, delta_sign: float,
+    ) -> None:
         """
         Recover the weights and biases of the first layer.
 
-        :param delta_init_value: Initial delta of weight value search
-        :type delta_init_value: `float`
-        :param delta_value_max: Maximum delta  of weight value search
-        :type delta_value_max: `float`
-        :param d2_min: Minimum acceptable value of sum of absolute second derivatives
-        :type d2_min: `float`
-        :param d_step:  Step size of delta increase
-        :type d_step: `float`
-        :param delta_sign: Delta of weight sign search
-        :type delta_sign: `float`
+        :param delta_init_value: Initial delta of weight value search.
+        :param delta_value_max: Maximum delta  of weight value search.
+        :param d2_min: Minimum acceptable value of sum of absolute second derivatives.
+        :param d_step:  Step size of delta increase.
+        :param delta_sign: Delta of weight sign search.
         """
         logger.info("Recovering weights of first layer.")
 
-        # Absolute Value Recovery
+        if self.num_neurons is None:
+            raise ValueError("The value of `num_neurons` is required for critical point search.")
 
+        # Absolute Value Recovery
         d2_ol_d2ej_xi = np.zeros((self.num_features, self.num_neurons), dtype=NUMPY_DTYPE)
 
         for i in range(self.num_neurons):
@@ -341,17 +324,18 @@ class FunctionallyEquivalentExtraction(ExtractionAttack):
                 if co_m < co_p * np.max(1 / self.a0_pairwise_ratios[:, i]):
                     self.a0_pairwise_ratios[j, i] *= -1
 
-    def _sign_recovery(self, unit_vector_scale):
+    def _sign_recovery(self, unit_vector_scale: int) -> None:
         """
         Recover the sign of weights in the first layer.
 
         :param unit_vector_scale: Multiplicative scale of the unit vector e_j.
-        :type unit_vector_scale: `int`
         """
         logger.info("Recover sign of the weights of the first layer.")
 
-        a0_pairwise_ratios_inverse = 1.0 / self.a0_pairwise_ratios
+        if self.num_neurons is None:
+            raise ValueError("The value of `num_neurons` is required for critical point search.")
 
+        a0_pairwise_ratios_inverse = 1.0 / self.a0_pairwise_ratios
         self.b_0 = np.zeros((self.num_neurons, 1), dtype=NUMPY_DTYPE)
 
         for i in range(self.num_neurons):
@@ -366,7 +350,6 @@ class FunctionallyEquivalentExtraction(ExtractionAttack):
         result_z = least_squares(f_z, z_0)
 
         for i in range(self.num_neurons):
-
             e_i = np.zeros((self.num_neurons, 1), dtype=NUMPY_DTYPE)
             e_i[i, 0] = unit_vector_scale
 
@@ -375,9 +358,7 @@ class FunctionallyEquivalentExtraction(ExtractionAttack):
                 return np.squeeze(np.matmul(-a0_pairwise_ratios_inverse.T, np.expand_dims(v_i, axis=0).T) - e_i)
 
             v_0 = np.random.normal(0, 1, self.num_features)
-
             result_v_i = least_squares(f_v, v_0)
-
             value_p = np.sum(
                 np.abs(
                     self._o_l(np.expand_dims(result_z.x, axis=0))
@@ -397,17 +378,18 @@ class FunctionallyEquivalentExtraction(ExtractionAttack):
 
         self.w_0 = a0_pairwise_ratios_inverse
 
-    def _last_layer_extraction(self, x):
+    def _last_layer_extraction(self, x: np.ndarray) -> None:
         """
         Extract weights and biases of the second layer.
 
-        :param x: Samples of input data of shape (num_samples, num_features).
-        :type x: `np.ndarray`
+        :param x: Samples of input data of shape `(num_samples, num_features)`.
         """
         logger.info("Extract second layer.")
 
-        predictions = self._o_l(x)
+        if self.num_neurons is None:
+            raise ValueError("The value of `num_neurons` is required for critical point search.")
 
+        predictions = self._o_l(x)
         w_1_b_1_0 = np.random.normal(0, 1, ((self.num_neurons + 1) * self.num_classes)).astype(dtype=NUMPY_DTYPE)
 
         def f_w_1_b_1(w_1_b_1_i):
@@ -431,15 +413,14 @@ if __name__ == "__main__":
     import tensorflow as tf
 
     tf.compat.v1.disable_eager_execution()
+    tf.keras.backend.set_floatx("float64")
 
     from tensorflow.keras.datasets import mnist
     from tensorflow.keras.models import Sequential
     from tensorflow.keras.layers import Dense
 
-    np.random.seed(0)
-
+    np.random.seed(1)
     number_neurons = 16
-
     batch_size = 128
     number_classes = 10
     epochs = 10
@@ -448,10 +429,8 @@ if __name__ == "__main__":
     number_channels = 1
 
     (x_train, y_train), (x_test, y_test) = mnist.load_data()
-
     x_train = x_train.reshape(x_train.shape[0], img_rows, img_cols, number_channels)
     x_test = x_test.reshape(x_test.shape[0], img_rows, img_cols, number_channels)
-
     input_shape = (number_channels * img_rows * img_cols,)
 
     x_train = x_train.reshape((x_train.shape[0], number_channels * img_rows * img_cols)).astype("float64")
@@ -479,7 +458,9 @@ if __name__ == "__main__":
             metrics=["accuracy"],
         )
 
-        model.fit(x_train, y_train, batch_size=batch_size, epochs=epochs, verbose=1, validation_data=(x_test, y_test))
+        model.fit(
+            x_train, y_train, batch_size=batch_size, epochs=epochs, verbose=1, validation_data=(x_test, y_test),
+        )
 
         model.save("./model.h5")
 
@@ -487,7 +468,9 @@ if __name__ == "__main__":
 
     target_classifier = KerasClassifier(model=model, use_logits=True, clip_values=(0, 1))
 
-    fee = FunctionallyEquivalentExtraction(classifier=target_classifier, num_neurons=number_neurons)
+    fee = FunctionallyEquivalentExtraction(
+        classifier=target_classifier, num_neurons=number_neurons  # type: ignore
+    )
     bbc = fee.extract(x_test[0:100])
 
     y_test_predicted_extracted = bbc.predict(x_test)
